import hashlib
from collections import defaultdict
from json import JSONDecodeError
from typing import TYPE_CHECKING

from httpx import HTTPStatusError, RequestError, Response
from tenacity import retry, retry_if_exception_type, stop_after_attempt

from maimai_py.exceptions import InvalidJsonError, MaimaiPyError

from .base import IAliasProvider

if TYPE_CHECKING:
    from maimai_py.maimai import MaimaiClient


class YuzuProvider(IAliasProvider):
    """The provider that fetches song aliases from the Yuzu.

    Yuzu is a bot API that provides song aliases for maimai DX.

    Yuzu: https://bot.yuzuchan.moe/
    """

    base_url = "https://www.yuzuchan.moe/api/"
    """The base URL for the Yuzu API."""

    def _hash(self) -> str:
        return hashlib.md5(b"yuzu").hexdigest()

    def _check_response(self, resp: Response) -> dict:
        try:
            resp_json = resp.json()
            if not resp.is_success:
                resp.raise_for_status()
            return resp_json
        except JSONDecodeError as exc:
            raise InvalidJsonError(resp.text) from exc
        except HTTPStatusError as exc:
            raise MaimaiPyError(exc) from exc

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_aliases(self, client: "MaimaiClient") -> dict[int, list[str]]:
        resp = await client._client.get(self.base_url + "maimaidx/maimaidxalias")
        resp_json = self._check_response(resp)
        grouped = defaultdict(list)
        for item in resp_json["content"]:
            grouped[item["SongID"] % 10000].extend(item["Alias"])
        return grouped



import hashlib
from datetime import datetime
from typing import TYPE_CHECKING

from httpcore import NetworkError, TimeoutException
from maimai_ffi import arcade
from tenacity import retry, retry_if_exception_type, stop_after_attempt

from maimai_py.models import *
from maimai_py.utils import ScoreCoefficient

from .base import IPlayerIdentifierProvider, IPlayerProvider, IRegionProvider, IScoreProvider

if TYPE_CHECKING:
    from maimai_py.maimai import MaimaiClient, MaimaiSongs


class ArcadeProvider(IPlayerProvider, IScoreProvider, IRegionProvider, IPlayerIdentifierProvider):
    """The provider that fetches data from the wahlap maimai arcade.

    This part of the maimai.py is not open-source, we distribute the compiled version of this part of the code as maimai_ffi.

    Feel free to ask us to solve if your platform or architecture is not supported.

    maimai.ffi: https://pypi.org/project/maimai-ffi
    """

    _http_proxy: Optional[str] = None

    def __init__(self, http_proxy: Optional[str] = None):
        self._http_proxy = http_proxy

    def _hash(self) -> str:
        return hashlib.md5(b"arcade").hexdigest()

    @staticmethod
    async def _deser_score(score: dict, songs: "MaimaiSongs") -> Optional[Score]:
        if song := await songs.by_id(score["musicId"] % 10000):
            song_type = SongType._from_id(score["musicId"])
            level_index = LevelIndex(score["level"]) if song_type != SongType.UTAGE else score["musicId"]
            achievement = float(score["achievement"]) / 10000
            if diff := song.get_difficulty(song_type, level_index):
                fs_type = FSType(score["syncStatus"]) if 0 < score["syncStatus"] < 5 else None
                fs_type = FSType.SYNC if score["syncStatus"] == 5 else fs_type
                return Score(
                    id=score["musicId"] if score["musicId"] > 100000 else score["musicId"] % 10000,
                    level=diff.level,
                    level_index=diff.level_index,
                    achievements=achievement,
                    fc=FCType(4 - score["comboStatus"]) if score["comboStatus"] != 0 else None,
                    fs=fs_type,
                    dx_score=score["deluxscoreMax"],
                    dx_rating=ScoreCoefficient(achievement).ra(diff.level_value),
                    play_count=score["playCount"],
                    play_time=None,
                    rate=RateType._from_achievement(achievement),
                    type=song_type,
                )

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(TitleServerNetworkError), reraise=True)
    async def get_player(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> ArcadePlayer:
        maimai_icons = await client.items(PlayerIcon)
        maimai_trophies = await client.items(PlayerTrophy)
        maimai_nameplates = await client.items(PlayerNamePlate)
        if identifier.credentials and isinstance(identifier.credentials, str):
            resp_dict = await arcade.get_user_preview(identifier.credentials.encode(), http_proxy=self._http_proxy)
            return ArcadePlayer(
                name=resp_dict["userName"],
                rating=resp_dict["playerRating"],
                is_login=resp_dict["isLogin"],
                name_plate=await maimai_nameplates.by_id(resp_dict["nameplateId"]),
                icon=await maimai_icons.by_id(resp_dict["iconId"]),
                trophy=await maimai_trophies.by_id(resp_dict["trophyId"]),
            )
        raise InvalidPlayerIdentifierError("Player identifier credentials should be provided.")

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(TitleServerNetworkError), reraise=True)
    async def get_scores_all(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[Score]:
        maimai_songs = await client.songs()
        if identifier.credentials and isinstance(identifier.credentials, str):
            resp_list = await arcade.get_user_scores(identifier.credentials.encode(), http_proxy=self._http_proxy)
            v = [s for score in resp_list if (s := await ArcadeProvider._deser_score(score, maimai_songs))]
            return v
        raise InvalidPlayerIdentifierError("Player identifier credentials should be provided.")

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(TitleServerNetworkError), reraise=True)
    async def get_regions(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[PlayerRegion]:
        if identifier.credentials and isinstance(identifier.credentials, str):
            resp_dict = await arcade.get_user_region(identifier.credentials.encode(), http_proxy=self._http_proxy)
            return [
                PlayerRegion(
                    region_id=region["regionId"],
                    region_name=region["regionName"],
                    play_count=region["playCount"],
                    created_at=datetime.strptime(region["created"], "%Y-%m-%d %H:%M:%S"),
                )
                for region in resp_dict["userRegionList"]
            ]
        raise InvalidPlayerIdentifierError("Player identifier credentials should be provided.")

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type((TimeoutException, NetworkError)), reraise=True)
    async def get_identifier(self, code: Union[str, dict[str, str]], client: "MaimaiClient") -> PlayerIdentifier:
        resp_bytes: bytes = await arcade.get_uid_encrypted(str(code), http_proxy=self._http_proxy)
        return PlayerIdentifier(credentials=resp_bytes.decode())



import hashlib
import typing
from json import JSONDecodeError
from typing import TYPE_CHECKING, Generator, Iterable

from httpx import HTTPStatusError, RequestError, Response
from tenacity import retry, retry_if_exception_type, stop_after_attempt

from maimai_py.models import *
from maimai_py.models import PlayerIdentifier, Score, Song

from .base import ICurveProvider, IPlayerProvider, IScoreProvider, IScoreUpdateProvider, ISongProvider

if TYPE_CHECKING:
    from maimai_py.maimai import MaimaiClient, MaimaiSongs


class DivingFishProvider(ISongProvider, IPlayerProvider, IScoreProvider, IScoreUpdateProvider, ICurveProvider):
    """The provider that fetches data from the Diving Fish.

    DivingFish: https://www.diving-fish.com/maimaidx/prober/
    """

    developer_token: Optional[str]
    """The developer token used to access the Diving Fish API."""
    base_url = "https://www.diving-fish.com/api/maimaidxprober/"
    """The base URL for the Diving Fish API."""

    @property
    def headers(self):
        """@private"""
        if not self.developer_token:
            raise InvalidDeveloperTokenError("Developer token is not provided.")
        return {"developer-token": self.developer_token}

    def __init__(self, developer_token: Optional[str] = None):
        """Initializes the DivingFishProvider.

        Args:
            developer_token: The developer token used to access the Diving Fish API.
        """
        self.developer_token = developer_token

    def _hash(self) -> str:
        return hashlib.md5(b"divingfish").hexdigest()

    @staticmethod
    def _deser_song(song: dict) -> Song:
        return Song(
            id=int(song["id"]) % 10000,
            title=song["basic_info"]["title"] if int(song["id"]) != 383 else "Link",
            artist=song["basic_info"]["artist"],
            genre=name_to_genre[song["basic_info"]["genre"]],
            bpm=song["basic_info"]["bpm"],
            map=None,
            rights=None,
            aliases=None,
            version=divingfish_to_version[song["basic_info"]["from"]].value,
            disabled=False,
            difficulties=SongDifficulties(standard=[], dx=[], utage=[]),
        )

    @staticmethod
    def _deser_diffs(song: dict) -> Generator[SongDifficulty, None, None]:
        song_id, song_type = int(song["id"]), SongType._from_id(song["id"])
        if song_type == SongType.STANDARD or song_type == SongType.DX:
            for idx, chart in enumerate(song["charts"]):
                yield SongDifficulty(
                    type=song_type,
                    level=song["level"][idx],
                    level_value=song["ds"][idx],
                    level_index=LevelIndex(idx),
                    note_designer=chart["charter"],
                    version=divingfish_to_version[song["basic_info"]["from"]].value,
                    tap_num=chart["notes"][0],
                    hold_num=chart["notes"][1],
                    slide_num=chart["notes"][2],
                    touch_num=chart["notes"][3] if song_type == SongType.DX else 0,
                    break_num=chart["notes"][4] if song_type == SongType.DX else chart["notes"][3],
                    curve=None,
                )
        elif song_type == SongType.UTAGE and len(song["charts"]) > 0:
            first_diff = song["charts"][0]
            second_diff = song["charts"][1] if len(song["charts"]) > 1 else None
            yield SongDifficultyUtage(
                diff_id=song_id,
                kanji=song["basic_info"]["title"][1:2],
                description="LET'S PARTY!",
                is_buddy=len(song["charts"]) == 2,
                type=song_type,
                level=song["level"][0],
                level_value=song["ds"][0],
                level_index=LevelIndex(0),
                note_designer=first_diff["charter"],
                version=divingfish_to_version[song["basic_info"]["from"]].value,
                tap_num=first_diff["notes"][0] + (second_diff["notes"][0] if second_diff else 0),
                hold_num=first_diff["notes"][1] + (second_diff["notes"][1] if second_diff else 0),
                slide_num=first_diff["notes"][2] + (second_diff["notes"][2] if second_diff else 0),
                touch_num=first_diff["notes"][3] + (second_diff["notes"][3] if second_diff else 0),
                break_num=first_diff["notes"][4] + (second_diff["notes"][4] if second_diff else 0),
                curve=None,
            )

    @staticmethod
    def _deser_score(score: dict) -> Score:
        return Score(
            id=score["song_id"] if score["song_id"] > 100000 else score["song_id"] % 10000,
            level=score["level"],
            level_index=LevelIndex(score["level_index"]),
            achievements=score["achievements"],
            fc=FCType[score["fc"].upper()] if score["fc"] else None,
            fs=FSType[score["fs"].upper()] if score["fs"] else None,
            dx_score=score["dxScore"],
            dx_rating=score["ra"],
            play_count=None,
            play_time=None,
            rate=RateType[score["rate"].upper()],
            type=SongType._from_id(score["song_id"]),
        )

    @staticmethod
    async def _ser_score(score: Score, songs: "MaimaiSongs") -> Optional[dict]:
        if song := await songs.by_id(score.id % 10000):
            song_title = "Link(CoF)" if score.id == 383 else song.title
            if score.type == SongType.UTAGE and (diff := song.get_difficulty(score.type, score.id)):
                diff = typing.cast(SongDifficultyUtage, diff)
                song_title = f"[{diff.kanji}]{song_title}"
            return {
                "title": song_title,
                "level_index": score.level_index.value,
                "achievements": score.achievements,
                "fc": score.fc.name.lower() if score.fc else None,
                "fs": score.fs.name.lower() if score.fs else None,
                "dxScore": score.dx_score,
                "type": score.type._to_abbr(),
            }

    @staticmethod
    def _deser_curve(chart: dict) -> CurveObject:
        return CurveObject(
            sample_size=int(chart["cnt"]),
            fit_level_value=chart["fit_diff"],
            avg_achievements=chart["avg"],
            stdev_achievements=chart["std_dev"],
            avg_dx_score=chart["avg_dx"],
            rate_sample_size={v: chart["dist"][13 - i] for i, v in enumerate(RateType)},
            fc_sample_size={v: chart["dist"][4 - i] for i, v in enumerate(FCType)},
        )

    def _check_response_player(self, resp: Response) -> dict:
        try:
            resp_json = resp.json()
            if resp.status_code in [400, 401]:
                raise InvalidPlayerIdentifierError(resp_json["message"])
            elif resp.status_code == 403:
                raise PrivacyLimitationError(resp_json["message"])
            elif "msg" in resp_json and resp_json["msg"] in ["请先联系水鱼申请开发者token", "开发者token有误", "开发者token被禁用"]:
                raise InvalidDeveloperTokenError(resp_json["msg"])
            elif "message" in resp_json and resp_json["message"] in ["导入token有误", "尚未登录", "会话过期"]:
                raise InvalidPlayerIdentifierError(resp_json["message"])
            elif not resp.is_success:
                resp.raise_for_status()
            return resp_json
        except JSONDecodeError as exc:
            raise InvalidJsonError(resp.text) from exc
        except HTTPStatusError as exc:
            raise MaimaiPyError(exc) from exc

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_songs(self, client: "MaimaiClient") -> list[Song]:
        resp = await client._client.get(self.base_url + "music_data")
        resp.raise_for_status()
        resp_json = resp.json()
        songs_unique: dict[int, Song] = {}
        for song in resp_json:
            song_key = int(song["id"]) % 10000
            song_type: SongType = SongType._from_id(song["id"])
            if song_key not in songs_unique:
                songs_unique[song_key] = DivingFishProvider._deser_song(song)
            difficulties: list[SongDifficulty] = songs_unique[song_key].difficulties.__getattribute__(song_type.value)
            difficulties.extend(DivingFishProvider._deser_diffs(song))
        return list(songs_unique.values())

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_player(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> DivingFishPlayer:
        resp = await client._client.post(self.base_url + "query/player", json=identifier._as_diving_fish())
        resp_json = self._check_response_player(resp)
        return DivingFishPlayer(
            name=resp_json["username"],
            rating=resp_json["rating"],
            nickname=resp_json["nickname"],
            plate=resp_json["plate"],
            additional_rating=resp_json["additional_rating"],
        )

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_scores_all(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[Score]:
        resp = await client._client.get(self.base_url + "dev/player/records", params=identifier._as_diving_fish(), headers=self.headers)
        resp_json = self._check_response_player(resp)
        return [s for score in resp_json["records"] if (s := DivingFishProvider._deser_score(score))]

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_scores_best(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[Score]:
        resp = await client._client.post(self.base_url + "query/player", json={"b50": True, **identifier._as_diving_fish()})
        resp_json = self._check_response_player(resp)
        return [DivingFishProvider._deser_score(score) for score in resp_json["charts"]["sd"] + resp_json["charts"]["dx"]]

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_scores_one(self, identifier: PlayerIdentifier, song: Song, client: "MaimaiClient") -> list[Score]:
        resp = await client._client.post(
            self.base_url + "dev/player/record",
            json={"music_id": list(song.get_divingfish_ids()), **identifier._as_diving_fish()},
            headers=self.headers,
        )
        resp_json: dict[str, dict] = self._check_response_player(resp)
        return [s for scores in resp_json.values() for score in scores if (s := DivingFishProvider._deser_score(score))]

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def update_scores(self, identifier: PlayerIdentifier, scores: Iterable[Score], client: "MaimaiClient") -> None:
        headers, cookies = None, None
        maimai_songs = await client.songs()
        if identifier.username and identifier.credentials:
            login_json = {"username": identifier.username, "password": identifier.credentials}
            resp1 = await client._client.post("https://www.diving-fish.com/api/maimaidxprober/login", json=login_json)
            self._check_response_player(resp1)
            cookies = resp1.cookies
        elif not identifier.username and identifier.credentials and isinstance(identifier.credentials, str):
            headers = {"Import-Token": identifier.credentials}
        else:
            raise InvalidPlayerIdentifierError("Either username and password or import token is required to deliver scores")
        scores_json = [json for score in scores if (json := await DivingFishProvider._ser_score(score, maimai_songs))]
        resp2 = await client._client.post(self.base_url + "player/update_records", cookies=cookies, headers=headers, json=scores_json)
        self._check_response_player(resp2)

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_curves(self, client: "MaimaiClient") -> dict[tuple[int, SongType], list[CurveObject]]:
        resp = await client._client.get(self.base_url + "chart_stats")
        resp.raise_for_status()
        return {
            (int(idx) % 10000, SongType._from_id(int(idx))): ([DivingFishProvider._deser_curve(chart) for chart in charts if chart != {}])
            for idx, charts in (resp.json())["charts"].items()
        }



import asyncio
import functools
import hashlib
import operator
import random
from typing import TYPE_CHECKING

from httpx import Cookies, RequestError
from tenacity import retry, retry_if_exception_type, stop_after_attempt

from maimai_py.models import *
from maimai_py.utils import HTMLScore, ScoreCoefficient, wmdx_html2score
from maimai_py.utils.page_parser import wmdx_html2player, wmdx_html2players, wmdx_html2record

from .base import IPlayerIdentifierProvider, IPlayerProvider, IRecordProvider, IScoreProvider

if TYPE_CHECKING:
    from maimai_py.maimai import MaimaiClient, MaimaiSongs


class WechatProvider(IScoreProvider, IPlayerProvider, IPlayerIdentifierProvider, IRecordProvider):
    """The provider that fetches data from the Wahlap Wechat OffiAccount.

    PlayerIdentifier must have the `credentials` attribute, we suggest you to use the `maimai.wechat()` method to get the identifier.

    PlayerIdentifier should not be cached or stored in the database, as the cookies may expire at any time.

    Wahlap Wechat OffiAccount: https://maimai.wahlap.com/maimai-mobile/
    """

    def _hash(self) -> str:
        return hashlib.md5(b"wechat").hexdigest()

    @staticmethod
    async def _deser_score(score: HTMLScore, songs: "MaimaiSongs") -> Optional[Score]:
        if song := await songs.by_title(score.title):
            is_utage = (len(song.difficulties.dx) + len(song.difficulties.standard)) == 0
            song_type = SongType.STANDARD if score.type == "SD" else SongType.DX if score.type == "DX" and not is_utage else SongType.UTAGE
            level_index = LevelIndex(score.level_index)
            if diff := song.get_difficulty(song_type, level_index):
                rating = ScoreCoefficient(score.achievements).ra(diff.level_value)
                return Score(
                    id=song.id,
                    level=diff.level,
                    level_index=level_index,
                    achievements=score.achievements,
                    fc=FCType[score.fc.upper()] if score.fc else None,
                    fs=FSType[score.fs.upper().replace("FDX", "FSD")] if score.fs else None,
                    dx_score=score.dx_score,
                    dx_rating=rating,
                    play_count=None,
                    play_time=score.play_time,
                    rate=RateType[score.rate.upper()],
                    type=song_type,
                )

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def _crawl_scores_diff(self, client: "MaimaiClient", diff: int, cookies: Cookies, maimai_songs: "MaimaiSongs") -> list[Score]:
        await asyncio.sleep(random.randint(0, 300) / 1000)  # sleep for a random amount of time between 0 and 300ms
        resp1 = await client._client.get(f"https://maimai.wahlap.com/maimai-mobile/record/musicGenre/search/?genre=99&diff={diff}", cookies=cookies)
        if resp1.status_code == 302:
            raise InvalidPlayerIdentifierError("Failed to fetch scores, possibly due to invalid cookies or maintenance.")
        scores = wmdx_html2score(str(resp1.text))
        return [r for score in scores if (r := await WechatProvider._deser_score(score, maimai_songs))]

    async def _crawl_scores(self, client: "MaimaiClient", cookies: Cookies, maimai_songs: "MaimaiSongs") -> Sequence[Score]:
        tasks = [asyncio.create_task(self._crawl_scores_diff(client, diff, cookies, maimai_songs)) for diff in [0, 1, 2, 3, 4]]
        results = await asyncio.gather(*tasks)
        return functools.reduce(operator.concat, results, [])

    async def get_scores_all(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[Score]:
        maimai_songs = await client.songs()  # Ensure songs are loaded in cache
        if not identifier.credentials or not isinstance(identifier.credentials, Cookies):
            raise InvalidPlayerIdentifierError("Wahlap wechat cookies are required to fetch scores")
        scores = await self._crawl_scores(client, identifier.credentials, maimai_songs)
        return list(scores)

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_player(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> WechatPlayer:
        trophies = await client.items(PlayerTrophy)
        if not identifier.credentials or not isinstance(identifier.credentials, Cookies):
            raise InvalidPlayerIdentifierError("Wahlap wechat cookies are required to fetch player information")
        resp = await client._client.get("https://maimai.wahlap.com/maimai-mobile/friend/userFriendCode/", cookies=identifier.credentials)
        if resp.status_code == 302:
            raise InvalidPlayerIdentifierError("Failed to fetch player information, possibly due to invalid cookies or maintenance.")
        player = wmdx_html2player(str(resp.text))
        return WechatPlayer(
            name=player.name,
            rating=player.rating,
            friend_code=player.friend_code,
            star=player.star,
            trophy=trophy[0] if (trophy := await trophies.filter(name=player.trophy_text)) else None,
            token=None,
        )

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_identifier(self, code: Union[str, dict[str, str]], client: "MaimaiClient") -> PlayerIdentifier:
        if isinstance(code, dict) and all([code.get("r"), code.get("t"), code.get("code"), code.get("state")]):
            headers = {
                "User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.138 Safari/537.36 NetType/WIFI MicroMessenger/7.0.20.1781(0x6700143B) WindowsWechat(0x6307001e)",
                "Host": "tgk-wcaime.wahlap.com",
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            }
            resp = await client._client.get("https://tgk-wcaime.wahlap.com/wc_auth/oauth/callback/maimai-dx", params=code, headers=headers)
            if resp.status_code == 302 and resp.next_request:
                resp_next = await client._client.get(resp.next_request.url, headers=headers)
                return PlayerIdentifier(credentials=resp_next.cookies)
            else:
                raise InvalidWechatTokenError("Invalid or expired Wechat token")
        raise InvalidWechatTokenError("Invalid Wechat token format, expected a dict with 'r', 't', 'code', and 'state' keys")

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_friends(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[WechatPlayer]:
        trophies = await client.items(PlayerTrophy)
        if not identifier.credentials or not isinstance(identifier.credentials, Cookies):
            raise InvalidPlayerIdentifierError("Wahlap wechat cookies are required to fetch friends")
        # Prefetch the first page of friends, in order to get the total number of friends
        resp = await client._client.get("https://maimai.wahlap.com/maimai-mobile/friend/", cookies=identifier.credentials)
        if resp.status_code == 302:
            raise InvalidPlayerIdentifierError("Failed to fetch friends, possibly due to invalid cookies or maintenance.")
        friend_num, players = wmdx_html2players(str(resp.text))
        # Fetch all friends by iterating through pages, 10 friends per page
        for page in range(2, (friend_num // 10) + 2):
            resp = await client._client.get(f"https://maimai.wahlap.com/maimai-mobile/friend/pages/?idx={page}", cookies=identifier.credentials)
            players.extend(wmdx_html2players(str(resp.text))[1])
        return [
            WechatPlayer(
                name=player.name,
                rating=player.rating,
                friend_code=player.friend_code,
                star=player.star,
                token=player.token,
                trophy=trophy[0] if (trophy := await trophies.filter(name=player.trophy_text)) else None,
            )
            for player in players
        ]

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def set_rival_on(self, identifier: PlayerIdentifier, rival: WechatPlayer, client: "MaimaiClient") -> None:
        if not identifier.credentials or not isinstance(identifier.credentials, Cookies):
            raise InvalidPlayerIdentifierError("Wahlap wechat cookies are required to set rival status")
        resp = await client._client.post(
            "https://maimai.wahlap.com/maimai-mobile/friend/rivalOn/",
            cookies=identifier.credentials,
            data={"idx": rival.friend_code, "token": rival.token},
        )
        resp.raise_for_status()

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def set_rival_off(self, identifier: PlayerIdentifier, rival: WechatPlayer, client: "MaimaiClient") -> None:
        if not identifier.credentials or not isinstance(identifier.credentials, Cookies):
            raise InvalidPlayerIdentifierError("Wahlap wechat cookies are required to set rival status")
        resp = await client._client.post(
            "https://maimai.wahlap.com/maimai-mobile/friend/rivalOff/",
            cookies=identifier.credentials,
            data={"idx": rival.friend_code, "token": rival.token},
        )
        resp.raise_for_status()

    @retry(stop=stop_after_attempt(3), retry=retry_if_exception_type(RequestError), reraise=True)
    async def get_records(self, identifier: PlayerIdentifier, client: "MaimaiClient") -> list[Score]:
        maimai_songs = await client.songs()  # Ensure songs are loaded in cache
        if not identifier.credentials or not isinstance(identifier.credentials, Cookies):
            raise InvalidPlayerIdentifierError("Wahlap wechat cookies are required to fetch scores")
        resp = await client._client.get(f"https://maimai.wahlap.com/maimai-mobile/record/", cookies=identifier.credentials)
        if resp.status_code == 302:
            raise InvalidPlayerIdentifierError("Failed to fetch records, possibly due to invalid cookies or maintenance.")
        scores = wmdx_html2record(str(resp.text))
        return [r for score in scores if (r := await WechatProvider._deser_score(score, maimai_songs))]



import re
from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from lxml import etree

link_dx_score = [372, 522, 942, 924, 1425]


@dataclass
class HTMLScore:
    __slots__ = ["title", "level", "level_index", "type", "achievements", "dx_score", "play_time", "rate", "fc", "fs", "ds"]
    title: str
    level: str
    level_index: int
    type: str
    achievements: float
    dx_score: int
    play_time: Optional[datetime]
    rate: str
    fc: str
    fs: str
    ds: int


@dataclass
class HTMLPlayer:
    __slots__ = ["name", "friend_code", "rating", "trophy_text", "trophy_rarity", "star", "token"]
    name: str
    friend_code: int
    rating: int
    star: int
    token: Optional[str]
    trophy_text: Optional[str]
    trophy_rarity: Optional[str]


def get_level_index(src: str) -> int:
    if src.find("remaster") != -1:
        return 4
    elif src.find("master") != -1:
        return 3
    elif src.find("expert") != -1:
        return 2
    elif src.find("advanced") != -1:
        return 1
    elif src.find("basic") != -1:
        return 0
    else:
        return -1


def get_music_icon(src: str) -> str:
    matched = re.search(r"((?:[dcbas]{1,3}|fc|ap|sync|fs|fdx)p?)(?:lus)?\.png", src)
    return matched.group(1) if matched else ""


def get_dx_score(element) -> tuple[int, int]:
    elem_text = "".join(element.itertext())

    parts = elem_text.strip().split("/")
    if len(parts) != 2:
        return (0, 0)

    try:
        score = int(parts[0].replace(" ", "").replace(",", ""))
        full_score = int(parts[1].replace(" ", "").replace(",", ""))
        return (score, full_score)
    except (ValueError, IndexError):
        return (0, 0)


def get_data_from_div(div) -> Optional[HTMLScore]:
    form = div.find(".//form")
    if form is None:
        return None

    # Find img element and get src attribute
    img = form.find(".//img")
    if img is None:
        return None

    img_src = img.get("src", "")

    # Determine type (SD or DX)
    if not re.search(r"diff_(.*).png", img_src):
        matched = re.search(r"music_(.*).png", img_src)
        type_ = "SD" if matched and matched.group(1) == "standard" else "DX"
    elif form.getparent().getparent().get("id") is not None:
        parent_id = form.getparent().getparent().get("id", "")
        type_ = "SD" if parent_id[:3] == "sta" else "DX"
    else:
        next_sibling = form.getparent().getnext()
        if next_sibling is not None:
            src = next_sibling.get("src", "")
            matched = re.search(r"_(.*).png", src)
            type_ = "SD" if matched and matched.group(1) == "standard" else "DX"
        else:
            type_ = "DX"  # Default

    # Extract data from form elements
    try:
        title_elem = form.xpath(".//div[contains(@class, 'music_name_block')]")
        level_elem = form.xpath(".//div[contains(@class, 'music_lv_block')]")
        score_elem = form.xpath(".//div[contains(@class, 'music_score_block')]")
        icon_elems = form.xpath(".//img[contains(@src, 'music_icon')]")
        level_index = get_level_index(img_src)

        score = get_score_from_elems(title_elem, level_elem, score_elem, icon_elems, level_index, type_)
        return score
    except (IndexError, AttributeError):
        return None


def get_score_from_elems(title_elem, level_elem: Optional[list], score_elem, icon_elems, level_index: int, type_: str) -> HTMLScore:
    # Use `.xpath("string(.)")` to get innerText (including all children's text), rather than the text in the element itself.
    title = title_elem[0].xpath("string(.)") if title_elem else ""
    if title != "\u3000":  # Corner case for id 1422 (如月车站)
        title = title.strip()
    level = level_elem[0].text.strip() if level_elem else ""

    achievements = float(score_elem[0].xpath("string(.)").strip()[:-1]) if score_elem else 0.0
    dx_score, full_dx_score = get_dx_score(score_elem[1] if score_elem else None)

    fs = fc = rate = ""
    if len(icon_elems) >= 3:
        fs = get_music_icon(icon_elems[0].get("src", ""))
        fc = get_music_icon(icon_elems[1].get("src", ""))
        rate = get_music_icon(icon_elems[2].get("src", ""))

    if title == "Link" and full_dx_score != link_dx_score[level_index]:
        title = "Link(CoF)"

    return HTMLScore(
        title=title,
        level=level,
        level_index=level_index,
        type=type_,
        achievements=achievements,
        dx_score=dx_score,
        play_time=None,
        rate=rate,
        fc=fc,
        fs=fs,
        ds=0,
    )


def wmdx_html2score(html: str) -> list[HTMLScore]:
    parser = etree.HTMLParser()
    root = etree.fromstring(html, parser)

    divs = root.xpath("//div[contains(@class, 'w_450') and contains(@class, 'm_15') and contains(@class, 'p_r') and contains(@class, 'f_0')]")

    results = []
    for div in divs:
        score = get_data_from_div(div)
        if score is not None:
            results.append(score)

    del parser, root, divs
    return results


def get_data_from_record_div(div) -> Optional[HTMLScore]:
    top, main = div.findall("./div")
    assert top.get("class").find("playlog_top_container") != -1
    main_class = re.match(r"playlog_(\w+)_container", main.get("class"))
    assert main_class is not None
    level_index = get_level_index(main_class.group(1))

    play_time_str = top.xpath(".//div[contains(@class, 'sub_title')]/span[2]")[0].text
    play_time = datetime.strptime(play_time_str, "%Y/%m/%d %H:%M")
    type_src = main.xpath(".//img[contains(@class, 'playlog_music_kind_icon')]")[0].get("src")
    matched = re.search(r"_(.*).png", type_src)
    type_ = "SD" if matched and matched.group(1) == "standard" else "DX"

    title_elem = main.xpath("./div[contains(@class, 'basic_block') and contains(@class, 'break')]")
    score_elem = main.xpath(".//div[contains(@class, 'playlog_achievement_txt')]") + main.xpath(".//div[contains(@class, 'playlog_score_block')]")
    achievement_elem = main.xpath(".//img[contains(@class, 'playlog_scorerank')]")
    icon_elems = main.xpath(".//img[contains(@class, 'h_35 m_5 f_l')]")[::-1] + achievement_elem

    score = get_score_from_elems(title_elem, None, score_elem, icon_elems, level_index, type_)
    score.play_time = play_time
    return score


def wmdx_html2record(html: str) -> list[HTMLScore]:
    parser = etree.HTMLParser()
    root = etree.fromstring(html, parser)

    divs = root.xpath("//div[contains(@class, 't_l') and contains(@class, 'v_b') and contains(@class, 'p_10') and contains(@class, 'f_0')]")

    results = []
    for div in divs:
        score = get_data_from_record_div(div)
        if score is not None:
            results.append(score)

    del parser, root, divs
    return results


def _extract_player_info(
    name_elements,
    friend_code_elements,
    rating_elements,
    trophy_elements,
    star_elements,
    token_elements=None,
    friend_code_is_input=False,
) -> HTMLPlayer:
    player_name = ""
    friend_code = 0
    rating = 0
    trophy_text = None
    trophy_rarity = None
    star = 0
    token = None

    if name_elements:
        player_name = name_elements[0].text.strip() if name_elements[0].text else ""

    if friend_code_elements:
        if friend_code_is_input:
            friend_code_text = friend_code_elements[0].get("value", "")
            if friend_code_text:
                friend_code = int(friend_code_text)
        else:
            friend_code_text = friend_code_elements[0].text.strip() if friend_code_elements[0].text else ""
            friend_code_numeric = re.sub(r"\D", "", friend_code_text)
            if friend_code_numeric:
                friend_code = int(friend_code_numeric)

    if rating_elements:
        rating_text = rating_elements[0].text.strip() if rating_elements[0].text else ""
        rating_numeric = re.sub(r"\D", "", rating_text)
        if rating_numeric:
            rating = int(rating_numeric)

    if trophy_elements:
        trophy_inner = trophy_elements[0]

        span_elements = trophy_inner.xpath(".//span")
        if span_elements:
            trophy_text = span_elements[0].text.strip() if span_elements[0].text else ""
        elif trophy_inner.text:
            trophy_text = trophy_inner.text.strip()

        trophy_block = trophy_inner.getparent()
        trophy_rarity = "Normal"  # Default rarity

        if trophy_block is not None:
            trophy_class = trophy_block.get("class", "")
            rarity_keywords = ["Rainbow", "Gold", "Silver", "Bronze", "Normal"]
            for rarity in rarity_keywords:
                if f"trophy_{rarity}" in trophy_class:
                    trophy_rarity = rarity
                    break

    if star_elements:
        star_text = star_elements[0].text.strip() if star_elements[0].text else ""
        star_match = re.search(r"×?(\d+)", star_text)
        if star_match:
            star = int(star_match.group(1))
        else:
            star_numeric = re.sub(r"\D", "", star_text)
            if star_numeric:
                star = int(star_numeric)

    if token_elements and len(token_elements) > 0:
        token = token_elements[0].get("value", None)

    return HTMLPlayer(
        name=player_name,
        friend_code=friend_code,
        rating=rating,
        trophy_text=trophy_text,
        trophy_rarity=trophy_rarity,
        star=star,
        token=token,
    )


def wmdx_html2player(html: str) -> HTMLPlayer:
    parser = etree.HTMLParser()
    root = etree.fromstring(html, parser)

    name_elements = root.xpath("//div[contains(@class, 'name_block') and contains(@class, 'f_l') and contains(@class, 'f_16')]")
    friend_code_elements = root.xpath(
        "//div[contains(@class, 'see_through_block') and contains(@class, 'm_t_5') and contains(@class, 'm_b_5') and contains(@class, 'p_5') and contains(@class, 't_c') and contains(@class, 'f_15')]"
    )
    rating_elements = root.xpath("//div[contains(@class, 'rating_block')]")
    trophy_elements = root.xpath("//div[contains(@class, 'trophy_inner_block') and contains(@class, 'f_13')]")
    star_elements = root.xpath("//div[contains(@class, 'p_l_10') and contains(@class, 'f_l') and contains(@class, 'f_14')]")

    player = _extract_player_info(name_elements, friend_code_elements, rating_elements, trophy_elements, star_elements)

    del parser, root, name_elements, friend_code_elements, rating_elements, trophy_elements, star_elements
    return player


def wmdx_html2players(html: str) -> tuple[int, list[HTMLPlayer]]:
    parser = etree.HTMLParser()
    root = etree.fromstring(html, parser)

    friend_count = 0
    friend_count_elems = root.xpath("//div[contains(@class, 'basic_block') and contains(text(), '好友数')]")
    if friend_count_elems:
        friend_count_text = "".join(friend_count_elems[0].itertext())
        match = re.search(r"好友数\s*\n?\s*(\d+)/\d+", friend_count_text)
        if match:
            friend_count = int(match.group(1))

    friend_divs = root.xpath(
        "//div[contains(@class, 'see_through_block') and contains(@class, 'p_r') and contains(@class, 'm_15') and contains(@class, 'm_t_5') and contains(@class, 'p_10') and contains(@class, 't_l') and contains(@class, 'f_0')]"
    )

    players = []
    for div in friend_divs:
        name_elements = div.xpath(".//div[contains(@class, 'name_block') and contains(@class, 'f_l') and contains(@class, 'f_16')]")
        friend_code_elements = div.xpath(".//input[@name='idx']")
        rating_elements = div.xpath(".//div[contains(@class, 'rating_block')]")
        trophy_elements = div.xpath(".//div[contains(@class, 'trophy_inner_block') and contains(@class, 'f_13')]")
        star_elements = div.xpath(".//div[contains(@class, 'p_l_10') and contains(@class, 'f_l') and contains(@class, 'f_14')]")
        token_elements = div.xpath(".//input[@name='token']")

        player = _extract_player_info(
            name_elements,
            friend_code_elements,
            rating_elements,
            trophy_elements,
            star_elements,
            token_elements,
            friend_code_is_input=True,
        )
        players.append(player)
        del name_elements, friend_code_elements, rating_elements, trophy_elements, star_elements, token_elements

    del parser, root, friend_count_elems, friend_divs
    return (friend_count, players)
